import logging
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Sequence

from rich.console import Console
from rich.text import Text

from ._app import OSXAPP
from ._commands import Command
from ._issues import Issue
from ._macho import sign_impl
from ._notarize import notarize_impl
from ._util import run_commands, serialize_to_sh


def configure_logging(verbose: int):
    verbosity = {0: logging.ERROR, 1: logging.WARNING, 2: logging.INFO, 3: logging.DEBUG}
    logging.basicConfig(format="%(levelname)s:%(message)s", level=verbosity[verbose])


def print_summary(app: OSXAPP, issues: Sequence[Issue]):
    console = Console()
    text = Text()
    text.append(f"Summary for {app.root}\n", style="bold magenta")
    if not issues:
        text.append("Found no issues!", style="bold lime")
    else:
        text.append(
            text.assemble(
                ("Found ", ""),
                (f"{len(issues)} ", "bold red"),
                ("issues of which ", ""),
                (f"{len([issue for issue in issues if issue.fixable])} ", "bold green"),
                ("can be fixed.", ""),
            )
        )

    console.print(text)


def print_unfixable(app: OSXAPP, issues: Sequence[Issue]):
    console = Console()
    text = Text()
    for issue in issues:
        text.append(
            text.assemble(
                ("Could ", ""),
                (f"not ", "bold red"),
                ("fix issue: ", ""),
                (f"{issue.details}\n ", "red"),
            )
        )

    console.print(text)


def parse_args() -> Namespace:

    common_args = ArgumentParser(add_help=False)
    common_args.add_argument("-v", "--verbose", action="count", default=0)
    common_args.add_argument("--no-progress", action="store_true")
    common_args.add_argument("--dry-run", action="store_true")
    common_args.add_argument("app_bundle", type=Path)
    common_args.add_argument("--sh-output", type=Path)

    fix_args = ArgumentParser(add_help=False)
    fix_args.add_argument("--rc-path-delete", action="store_true")
    fix_args.add_argument("--force-update", action="store_true")

    sign_args = ArgumentParser(add_help=False)
    sign_args.add_argument("entitlement_file", type=Path)
    sign_args.add_argument("developer_id", type=str)

    notarize_args = ArgumentParser(add_help=False)
    notarize_args.add_argument("-v", "--verbose", action="count", default=0)
    notarize_args.add_argument("app_bundle", type=Path)
    notarize_args.add_argument("keychain_profile", type=str)
    notarize_args.add_argument("keychain", type=Path)
    notarize_args.add_argument("apple_id_email", type=str)
    notarize_args.add_argument("team_id", type=str)

    parser = ArgumentParser()

    subparsers = parser.add_subparsers(dest="action", help="subcommand")
    check = subparsers.add_parser("check", parents=[common_args], help="Same as fix with '--dry-run'")

    fix = subparsers.add_parser(
        "fix",
        parents=[common_args, fix_args],
        help="Fix common issues that prevent your .app bundle from being notarized.",
    )
    sign = subparsers.add_parser(
        "sign",
        parents=[common_args, sign_args],
        help="Deep signing of binaries contained in your .app bundle (also within .jar archives) in a way that they will pass notarization/gatekeeper.",
    )

    fixsign = subparsers.add_parser(
        "fixsign", parents=[common_args, sign_args, fix_args], help="Do 'fix' and 'sign' in one go."
    )

    notarize = subparsers.add_parser(
        "notarize",
        parents=[notarize_args],
        help=(
            "Compress, submit, wait for notarization to complete and staple your app. "
            "This command does not offer a .sh out file option. "
        ),
    )

    return parser.parse_args()


def check(app: OSXAPP):
    """Check if .app bundle is likely to pass MacOs Gatekeeper

    Check integrity of binaries:
      * no RC_PATH outside the app
      * all linked libraries can be reached within the app or
        in "/System/", "/usr/", "/Library/".

    Check all binaries are signed.
    """
    return fix(app)


def fix(app: OSXAPP, rc_path_delete: bool = False, force_update: bool = False, need_repack=True) -> list[Command]:
    """Fix issues in mach-o libraries .app bundle

    Remove paths that point outside the app.

    Args:
        rc_path_delete: delete rc_paths that point outside the app. Use with care

    """
    issues = app.check_macho_binaries(rc_path_delete=rc_path_delete)
    issues.extend(app.check_jar_binaries(force_update=force_update))
    print_summary(app, issues)

    unfixable = [issue for issue in issues if not issue.fixable]
    print_unfixable(app, unfixable)

    commands: list[Command] = []
    for issue in issues:
        if issue.fixable:
            assert issue.fix is not None
            commands.append(issue.fix)

    if need_repack:
        commands.extend(app.jar_repack)
    return commands


def sign(app: OSXAPP, entitlement_file: Path, developer_id: str) -> list[Command]:
    # For jars, we need to sign and repack before signing  all
    assert entitlement_file.exists(), entitlement_file
    commands: list[Command] = []
    for jar in app.jars:
        commands.extend(jar.sign(entitlement_file, developer_id))

    # need to repack before signing the rest of the app
    commands.extend(app.jar_repack)

    for binary in app.macho_binaries:
        commands.append(sign_impl(entitlement_file, developer_id, binary.path))

    commands.append(sign_impl(entitlement_file, developer_id, app.bundle_exe))
    commands.append(sign_impl(entitlement_file, developer_id, app.root))

    return commands


def fixsign(
    app: OSXAPP, entitlement_file: Path, developer_id: str, rc_path_delete: bool = False, force_update: bool = False
):
    commands = fix(app, rc_path_delete, force_update, need_repack=False)
    commands.extend(sign(app, entitlement_file, developer_id))

    return commands


def notarize(app_path: Path, keychain_profile: str, keychain: Path, apple_id_email: str, team_id: str) -> int:
    success = notarize_impl(app_path, keychain_profile, keychain, apple_id_email, team_id)
    return success


def main():
    args = parse_args()
    configure_logging(verbose=args.verbose)

    commands: list[Command] = []
    match args.action:
        case "notarize":
            return notarize(args.app_bundle, args.keychain_profile, args.keychain, args.apple_id_email, args.team_id)
        case "check" | "fix" | "sign" | "fixsign":
            app = OSXAPP.from_path(args.app_bundle, with_progress=not args.no_progress)
            commands.extend(app.jar_extract)
            match args.action:
                case "check":
                    # force dry_run to be true for now
                    args.dry_run = True
                    commands.extend(check(app))
                case "fix":
                    commands.extend(fix(app, args.rc_path_delete, args.force_update))
                case "sign":
                    commands.extend(sign(app, args.entitlement_file, args.developer_id))
                case "fixsign":
                    commands.extend(
                        fixsign(app, args.entitlement_file, args.developer_id, args.rc_path_delete, args.force_update)
                    )
        case _:
            raise ValueError(f"Unexpected action {args.action}")

    if args.sh_output:
        serialize_to_sh(commands, args.sh_output)

    if not args.dry_run:
        run_commands(commands)


if __name__ == "__main__":
    sys.exit(main())



from pathlib import Path
from unittest.mock import Mock, patch

from app_pass._commands import Command
from app_pass._util import run_commands, serialize_to_sh


def test_command_serialization_to_sh_with_cwd():
    cmd = Command(args=["ls", "-l"], cwd=Path("/some/path"))

    sh_cmd = cmd.to_sh()

    assert len(sh_cmd) == 3
    assert sh_cmd[0] == '"cd" "/some/path"'
    assert sh_cmd[1] == '"ls" "-l"'
    assert sh_cmd[2] == '"cd" "-"'


def test_command_serialization_to_sh_with_cwd_with_comment():
    cmd = Command(args=["ls", "-l"], cwd=Path("/some/path"), comment="Something\nMultiline")

    sh_cmd = cmd.to_sh()

    assert len(sh_cmd) == 5
    assert sh_cmd[0] == "# Something"
    assert sh_cmd[1] == "# Multiline"
    assert sh_cmd[2] == '"cd" "/some/path"'
    assert sh_cmd[3] == '"ls" "-l"'
    assert sh_cmd[4] == '"cd" "-"'


def test_command_serialization_to_sh_no_cwd():
    cmd = Command(args=["ls", "-l"], cwd=None)

    sh_cmd = cmd.to_sh()

    assert len(sh_cmd) == 1
    assert sh_cmd[0] == '"ls" "-l"'


def test_serialize_to_sh(tmp_path: Path):
    cmds = [Command(["echo", f"{i}"]) for i in range(42)]

    sh_file = tmp_path / "sh_out.sh"

    serialize_to_sh(cmds, sh_file)

    lines = sh_file.read_text().split("\n")
    assert len(lines) == 42

    expected = [f'"echo" "{i}"' for i in range(42)]
    assert lines == expected


def test_serialize_to_sh_with_comment(tmp_path: Path):
    cmds = [Command(["echo", f"{i}"], comment=f"This echos {i}") for i in range(42)]

    sh_file = tmp_path / "sh_out.sh"

    serialize_to_sh(cmds, sh_file)

    lines = sh_file.read_text().split("\n")
    assert len(lines) == 42 * 2

    expected_comments = [f"# This echos {i}" for i in range(42)]
    expected_runs = [f'"echo" "{i}"' for i in range(42)]
    expected = [x for pair in zip(expected_comments, expected_runs) for x in pair]

    assert lines == expected


def test_serialize_to_sh_with_cwd(tmp_path: Path):
    cmds = [Command(["echo", f"{i}"], cwd=Path(f"/my/home{i}")) for i in range(42)]

    sh_file = tmp_path / "sh_out.sh"

    serialize_to_sh(cmds, sh_file)

    lines = sh_file.read_text().split("\n")
    assert len(lines) == 42 * 3

    expected_cd_in = [f'"cd" "/my/home{i}"' for i in range(42)]
    expected_runs = [f'"echo" "{i}"' for i in range(42)]
    expected_cd_out = ['"cd" "-"' for i in range(42)]
    expected = [x for triple in zip(expected_cd_in, expected_runs, expected_cd_out) for x in triple]

    assert lines == expected


def test_serialize_to_sh_with_all(tmp_path: Path):
    cmds = [Command(["echo", f"{i}"], cwd=Path(f"/my/home{i}"), comment=f"This echos {i}") for i in range(42)]

    sh_file = tmp_path / "sh_out.sh"

    serialize_to_sh(cmds, sh_file)

    lines = sh_file.read_text().split("\n")
    assert len(lines) == 42 * 4

    expected_comments = [f"# This echos {i}" for i in range(42)]
    expected_cd_in = [f'"cd" "/my/home{i}"' for i in range(42)]
    expected_runs = [f'"echo" "{i}"' for i in range(42)]
    expected_cd_out = ['"cd" "-"' for i in range(42)]
    expected = [x for quad in zip(expected_comments, expected_cd_in, expected_runs, expected_cd_out) for x in quad]

    assert lines == expected


def test_run_commands():
    cmds = [Command(["echo", f"{i}"], cwd=Path(f"/my/home{i}"), comment=f"This echos {i}") for i in range(42)]

    with patch("subprocess.run", new=Mock(return_value=Mock(returncode=0))) as popen_mock:
        run_commands(cmds)

    assert popen_mock.call_count == 42
    args = popen_mock.call_args_list
    for call, expected_call in zip(args, [["echo", f"{i}"] for i in range(42)]):
        assert call.args[0] == expected_call

    for call, expected_path in zip(args, [Path(f"/my/home{i}") for i in range(42)]):
        assert call.kwargs["cwd"] == expected_path



from dataclasses import dataclass
from pathlib import Path
from typing import Optional


@dataclass
class Command:
    # One or multiple commands that can be executed in the shell
    args: list[str]
    cwd: Optional[Path] = None
    comment: Optional[str] = None
    run_python: bool = True
    """
    Some commands are only added for the .sh output (e.g. creating temp directories)
    Those are marked with run_python=False
    """
    retry_backoff: bool = False
    """
    More or less a hunch that some commands can fail for various reasons.
    One such example is codesign, which needs to connect to the apple timestamp
    server.
    Mark those commands that something useful can be done ;/
    """

    def to_sh(self) -> list[str]:
        args = [f'"{arg}"' for arg in self.args]
        cmds = [" ".join(args)]
        if self.cwd:
            cmds.insert(0, f'"cd" "{self.cwd}"')
            cmds.append('"cd" "-"')

        if self.comment:
            # fix multi-line comments - who knows
            cmds = [f"# {c}" for c in self.comment.split("\n")] + cmds

        return cmds



import contextlib
import logging
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import List, Optional

from lxml import etree
from rich.progress import Progress

from app_pass._commands import Command

from ._issues import BuildIssue, Issue, LibraryPathIssue, RcpathIssue
from ._jar import Jar
from ._macho import Build, MachOBinary, fix_lib_id, fix_load_path, fix_rpath, parse_macho, remove_rpath, vtool_overwrite
from ._util import BinaryType, iter_all_binaries

logger = logging.getLogger(__name__)


def parse_plist(plist: Path):
    """partial read of plist"""
    doc = etree.parse(plist)
    root = doc.getroot()
    return {x.text: x.getnext().text for x in root.findall(".//dict/key")}


# for now the assumption is that whenever these are encountered, things should be fine.
_ALLOWED_SPECIAL = [Path(x) for x in ["@rpath", "@executable_path", "@loader_path"]]
_ALLOWED_SYSTEM = [Path(x) for x in ["/System/", "/usr/", "/Library/"]]


@dataclass
class OSXAPP:
    root: Path
    loader_path: Path  # dir
    bundle_exe: Path  # file
    macho_binaries: list[MachOBinary]
    jars: list[Jar]
    # TODO: make build configurable
    default_build: Build = Build(platform="macos", minos="10.10", sdk="10.10")

    @staticmethod
    def from_path(root: Path, with_progress=True) -> "OSXAPP":
        if not root.is_absolute():
            root = root.resolve()
        plist = root / "Contents" / "Info.plist"
        assert plist.exists()
        plist_dict = parse_plist(plist)
        # I've seen both, the executable "CFBundleExecutable" being only the `binary` within "MacOS" folder
        # and `MacOS/binary`:
        executable = Path(plist_dict["CFBundleExecutable"])
        if executable.parent == Path("MacOS"):
            bundle_exe = root / "Contents" / executable
        else:
            bundle_exe = root / "Contents" / "MacOS" / executable

        loader_path = bundle_exe.parent

        macho_binaries: list[MachOBinary] = []
        jars: list[Jar] = []

        if with_progress:
            prog = Progress
        else:
            prog = contextlib.nullcontext

        with prog() as progress:
            for f, bin_type in iter_all_binaries(root, progress):
                if bin_type == BinaryType.MACHO:
                    macho_bin = parse_macho(f)
                    macho_binaries.append(macho_bin)
                elif bin_type == BinaryType.JAR:
                    jars.append(Jar.from_path(f, progress))

        return OSXAPP(root, loader_path, bundle_exe, macho_binaries, jars)

    def __post_init__(self):
        assert self.bundle_exe.exists(), self.bundle_exe
        assert self.bundle_exe.is_file()
        assert self.bundle_exe.is_relative_to(self.root), self.bundle_exe

    @property
    def jar_extract(self) -> list[Command]:
        commands = []
        for jar in self.jars:
            if jar.binaries:
                commands.extend(jar.create_commands)
        return commands

    @property
    def jar_repack(self) -> list[Command]:
        commands = []
        for jar in self.jars:
            if jar.binaries:
                commands.extend(jar.repack())
        return commands

    @cached_property
    def libraries(self) -> dict[str, MachOBinary]:
        return {x.path.name: x for x in self.macho_binaries}

    def lib_loader_relative(self, libname):
        assert libname in self.libraries
        lib_path = self.libraries[libname].path

        loader_root_relative = self.loader_path.relative_to(self.root)

        candiates = [loader_root_relative] + list(loader_root_relative.parents)

        loader_relative_path = None
        for i, candidate in enumerate(candiates):
            if lib_path.is_relative_to(self.root / candidate):
                up = "/".join([".."] * i)
                loader_relative_path = Path("@loader_path") / up / lib_path.relative_to(self.root / candidate)
                break

        if loader_relative_path:
            return loader_relative_path
        else:
            raise ValueError(f"Could not determine loader relative path for {lib_path} in {self.root=}")

    @cached_property
    def bundle_exe_rpaths(self) -> List[Path]:
        filtered = list(filter(lambda x: x.path == self.bundle_exe, self.macho_binaries))
        assert len(filtered) == 1
        bundle_exe_macho = filtered[0]
        if any(rc.is_absolute() for rc in bundle_exe_macho.rpaths):
            raise ValueError(
                f"{bundle_exe_macho.rpaths=} in {bundle_exe_macho.path=} need fixing, may not be absolute."
            )
        return bundle_exe_macho.rpaths

    def check_macho_binaries(self, rc_path_delete: bool = False) -> List[Issue]:
        issues = []
        for macho_binary in self.macho_binaries:
            id_issues = check_id_needs_fix(self, macho_binary)
            issues.extend(id_issues)
            lib_issues = check_libs_need_fix(self, macho_binary)
            issues.extend(lib_issues)
            rc_issues = check_rpaths_need_fix(self, macho_binary, rc_path_delete)
            issues.extend(rc_issues)

            if macho_binary.build and not macho_binary.build.is_valid:
                if macho_binary.build.can_fix:
                    valid_build = macho_binary.build.valid_build(self.default_build)
                    issue = BuildIssue(
                        fixable=True,
                        details="Missing build number.",
                        fix=vtool_overwrite(macho_binary.path, valid_build),
                    )
                else:
                    issue = BuildIssue(
                        fixable=False,
                        details=(
                            "Probably sdk for build outdated - gatekeeper requires >=10.9 "
                            f"({macho_binary.path}: {macho_binary.build.platform=} {macho_binary.build.sdk=} {macho_binary.build.minos=}). "
                            "Solving this properly means rebuilding the binary with a deployment target 10.9 "
                            "(for x64) or 11.0 (for arm64). Alternatively you can try the --force-update flag, that "
                            "simply overwrites the property."
                        ),
                    )
                issues.append(issue)
                if issue.fixable:
                    logger.info(f"Issue found: {macho_binary.path}, {type(issue)}, {issue.fixable=}")
                else:
                    logger.warning(f"Issue found: {macho_binary.path}, {type(issue)}, {issue.fixable=}")

        return issues

    def check_jar_binaries(self, force_update: bool = False) -> List[Issue]:
        issues = []
        for jar in self.jars:
            for binary in jar.binaries:

                if binary.build and not binary.build.is_valid:
                    if binary.build.can_fix:
                        valid_build = binary.build.valid_build(self.default_build)
                        issue = BuildIssue(
                            fixable=True,
                            details=f"({binary.path} ({jar.path}): Missing build number.",
                            fix=vtool_overwrite(binary.path, valid_build),
                        )
                    elif force_update:
                        valid_build = binary.build.valid_build(self.default_build, overwrite=force_update)
                        issue = BuildIssue(
                            fixable=True,
                            details=f"({binary.path} ({jar.path}): Missing build number.",
                            fix=vtool_overwrite(binary.path, valid_build),
                        )
                    else:
                        issue = BuildIssue(
                            fixable=False,
                            details=(
                                "Probably sdk for build outdated - gatekeeper requires >=10.9 "
                                f"({binary.path} ({jar.path}): {binary.build.platform=} {binary.build.sdk=} {binary.build.minos=}). "
                                "Solving this properly means rebuilding the binary with a deployment target 10.9 "
                                "(for x64) or 11.0 (for arm64). Alternatively you can try the --force-update flag, that "
                                "simply overwrites the property."
                            ),
                        )
                    issues.append(issue)
                    if issue.fixable:
                        logger.info(f"Issue found: {binary.path}, {type(issue)}, {issue.fixable=}")
                    else:
                        logger.warning(f"Issue found: {binary.path}, {type(issue)}, {issue.fixable=}")

        return issues


def fix_path_pointer(app: OSXAPP, path: Path) -> Optional[Path]:
    """
    Return a modified, valid path inside the app for a given path
    """
    if any(path.is_relative_to(x) for x in _ALLOWED_SYSTEM + _ALLOWED_SPECIAL):
        # could still be broken but if the app is running at all, this should
        # be fine
        return path

    if path.is_absolute() and path.is_relative_to(app.root):
        # we should be able to fix it somehow.
        # check if relative to one of the rpaths of the main executable
        app_rpaths = app.bundle_exe_rpaths
        for rpath in app_rpaths:
            assert not rpath.is_absolute()
            if "@loader_path/" in str(rpath):
                rc_abs = app.loader_path / Path(str(rpath).replace("@loader_path/", ""))
            elif "@executable_path/" in str(rpath):
                rc_abs = app.loader_path / Path(str(rpath).replace("@executable_path/", ""))
            elif "@executable_path/" in str(rpath):
                raise ValueError("Didn't really expect a relative path with @rpath, but it seems to exist :).")
            else:
                raise ValueError(f"Could not resolve rc_path - probably not valid {path}")

            if path.is_relative_to(rc_abs):
                new_path = path.relative_to(rc_abs)
                return rpath / new_path

        # fallback to loader_path and hope for the best (will break if loaded
        # based on some other executable)
        path = path.relative_to(app.loader_path)
        return Path("@loader_path") / path


def check_id_needs_fix(app: OSXAPP, binary: MachOBinary) -> List[LibraryPathIssue]:
    if not binary.id_:
        return []

    if not any(binary.id_.is_relative_to(x) for x in _ALLOWED_SYSTEM + _ALLOWED_SPECIAL):
        return [
            LibraryPathIssue(
                fixable=True,
                details="Library ID with fixed path",
                fix=fix_lib_id(binary.path, Path("@rpath") / binary.id_.name),
            )
        ]
    else:
        return []


def check_libs_need_fix(app: OSXAPP, binary: MachOBinary) -> List[LibraryPathIssue]:
    invalid = []
    for lib in binary.dylibs:
        if not any(lib.is_relative_to(x) for x in _ALLOWED_SYSTEM + _ALLOWED_SPECIAL):
            invalid.append(lib)

    issues = []
    # try to find a direct hit for the library
    for lib in invalid:
        if lib.name in app.libraries:
            found = app.lib_loader_relative(lib.name)
            issues.append(
                LibraryPathIssue(
                    fixable=True,
                    details="Link to library not valid",
                    fix=fix_load_path(binary.path, lib, found),
                )
            )
        else:
            issues.append(LibraryPathIssue(fixable=False, details=f"Issue with {lib.name} at {lib} for {binary.path}"))

    return issues


def check_rpaths_need_fix(app: OSXAPP, binary: MachOBinary, rc_path_delete: bool) -> List[RcpathIssue]:
    issues: List[RcpathIssue] = []
    for pth in binary.rpaths:
        fixed = fix_path_pointer(app, pth)
        if fixed and fixed != pth:
            issues.append(
                RcpathIssue(
                    fixable=True,
                    details=f"Rcpath fix: {pth} -> {fixed}",
                    fix=fix_rpath(binary.path, pth, fixed),
                )
            )

        if not fixed:
            if rc_path_delete:
                issues.append(
                    RcpathIssue(
                        fixable=True,
                        details=(
                            f"DELETING rpath in {binary.path} pointing outside of binary and allowed system paths, "
                            f"this may indicate build issues {pth}."
                        ),
                        fix=remove_rpath(binary.path, pth),
                    )
                )
            else:
                issues.append(
                    RcpathIssue(
                        fixable=False,
                        details=(
                            f"rpath in {binary.path} pointing outside of binary and allowed system paths, "
                            f"this may indicate build issues {pth}. You can force removal of such paths  "
                            "with `--rc-path-delete`."
                        ),
                    )
                )

    return issues



import enum
import logging
import pathlib
import struct
import subprocess
import time
from dataclasses import dataclass
from typing import Iterator, Optional, Tuple

from rich.progress import Progress

from ._commands import Command

logger = logging.getLogger(__name__)


class BinaryType(enum.Enum):
    NONE = enum.auto()
    MACHO = enum.auto()
    JAR = enum.auto()


@dataclass
class BinaryObj:
    path: pathlib.Path


# kind of assuming little endian...
MACHOMAGIC = (
    0xFEEDFACF,
    0xFEEDFACE,
    0xBEBAFECA,
    # okay whatever, doesn't hurt to check the other ordering, too
    # false positives will be weeded out by `file` anyway
    0xCFFAEDFE,
    0xCEFAEDFE,
    0xCAFEBABE,
)


def run_logged(command: Command) -> str:
    logger.debug(f"Executing command {' '.join(command.args)}")

    out = subprocess.run(command.args, stdout=subprocess.PIPE, stdin=subprocess.PIPE, cwd=command.cwd)
    if out.returncode != 0:
        logger.warning(f"Nonzero exit code ({out.returncode}) from command {' '.join(command.args)}")
        raise subprocess.CalledProcessError(
            returncode=out.returncode,
            cmd=command.args,
            stderr=out.stderr.decode("utf-8") if out.stderr else "",
            output=out.stdout.decode("utf-8") if out.stdout else "",
        )

    logger.debug(f"Successful command {' '.join(command.args)}")

    return out.stdout.decode("utf-8")


def run_commands(commands: list[Command]):
    last_backoff = object()

    for command in commands:
        if not command.run_python:
            continue

        if command.retry_backoff:
            backoff = [10, 30, last_backoff]
        else:
            backoff = [last_backoff]

        for sleep_time in backoff:
            try:
                run_logged(command)
            except subprocess.CalledProcessError:
                if sleep_time == last_backoff:
                    raise
                logger.info(f"Retrying... after {sleep_time}s")
                time.sleep(sleep_time)
            else:
                break


def serialize_to_sh(commands: list[Command], sh_cmd_out: pathlib.Path):
    cmds = []
    for cmd in commands:
        cmds.extend(cmd.to_sh())
    if sh_cmd_out.exists():
        logger.warning(f"Found {sh_cmd_out} - overwriting.")

    sh_cmd_out.write_text("\n".join(cmds))


def is_binary(path: pathlib.Path) -> BinaryType:
    if path.is_dir():
        return BinaryType.NONE

    if path.suffix in (".a", ".o"):
        logger.debug(f"Ignoring .a, and .o files: {path}")
        return BinaryType.NONE

    if path.suffix in (".py", ".pyc", ".txt", ".md", ".class", ".cpp", ".hpp", ".cxx", ".hxx", ".c", ".h", ".class"):
        return BinaryType.NONE

    if path.suffix in (".jar", ".sym"):
        file_out = run_logged(Command(["file", str(path)])).lower()
        if "java archive data (jar)" in file_out or "zip archive data" in file_out:
            return BinaryType.JAR

    with open(path, "rb") as f:
        magic_bytes = f.read(4)

    if len(magic_bytes) != 4:
        return BinaryType.NONE

    xx = struct.unpack("<I", magic_bytes)

    if xx[0] in MACHOMAGIC:
        file_out = run_logged(Command(["file", str(path)])).lower()
        if "mach-o" in file_out:
            if "architectures" in file_out:
                logger.warning(f"Multiple architectures in file {path}")
            return BinaryType.MACHO

    return BinaryType.NONE


def iter_all_binaries(
    root: pathlib.Path,
    progress: Optional[Progress],
) -> Iterator[Tuple[pathlib.Path, BinaryType]]:
    files = list(root.glob("**/*"))

    task = None
    if progress is not None:
        task = progress.add_task("Scanning files", total=len(files))
    for f in files:
        if f.is_symlink():
            continue
        binary_type = is_binary(f)
        if binary_type != BinaryType.NONE:
            yield f, binary_type
        if progress is not None:
            assert task is not None
            progress.advance(task, 1)

    if progress is not None:
        assert task is not None
        progress.remove_task(task)



from dataclasses import dataclass
from typing import Optional

from ._commands import Command


@dataclass
class Issue:
    fixable: bool
    details: str
    fix: Optional[Command] = None


@dataclass
class BuildIssue(Issue):
    pass


@dataclass
class RcpathIssue(Issue):
    pass


@dataclass
class LibraryPathIssue(Issue):
    pass



import atexit
import logging
import shutil
import tempfile
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Optional

from rich.progress import Progress

from app_pass._commands import Command
from app_pass._macho import MachOBinary, parse_macho, sign_impl

from ._util import BinaryObj, BinaryType, is_binary, run_logged

logger = logging.getLogger(__name__)


@dataclass
class Jar(BinaryObj):

    temp_path: Path
    binaries: list[MachOBinary]

    @staticmethod
    def from_path(p: Path, progress: Optional[Progress]) -> "Jar":
        if progress:
            task = progress.add_task(f"tempdir({p.name})", total=None)
        t = tempfile.mkdtemp()
        run_logged(Command(["ditto", "-x", "-k", str(p), t]))

        files = list(Path(t).glob("**/*"))

        if progress:
            progress.update(task, total=len(files))

        machos = []
        for file in files:
            binary_type = is_binary(file)

            if binary_type == BinaryType.MACHO:
                machos.append(parse_macho(file))
            elif binary_type == BinaryType.JAR:
                logger.warning(f"Nested jar in {p}: {file} - not expected")

            if progress:
                progress.advance(task, 1)

        if progress:
            progress.remove_task(task)

        def _cleanup(t):
            print(f"Cleaning up {t}")
            shutil.rmtree(t, ignore_errors=True)

        atexit.register(partial(_cleanup, t))

        return Jar(path=p, temp_path=Path(t), binaries=machos)

    @property
    def create_commands(self) -> list[Command]:
        return [
            Command(["mkdir", "-p", str(self.temp_path)], run_python=False),
            Command(["ditto", "-x", "-k", str(self.path), str(self.temp_path)], run_python=False),
        ]

    def sign(self, entitlement_file, developer_id) -> list[Command]:
        # nothing needs to be done if there aren't any binaries
        if not self.binaries:
            return []

        sign_commands = []

        for binary in self.binaries:
            sign_commands.append(sign_impl(entitlement_file, developer_id, binary.path))

        return sign_commands

    def repack(self) -> list[Command]:
        pack_command = Command(
            args=["ditto", "-v", "-c", "-k", str(self.temp_path), self.path.with_suffix(".zip").name],
            cwd=self.temp_path,
        )
        move_command = Command(args=["mv", str(self.temp_path / self.path.with_suffix(".zip").name), str(self.path)])
        return [pack_command, move_command]



import pathlib
from unittest.mock import Mock, PropertyMock, patch

import pytest

from app_pass._util import BinaryType, is_binary


@pytest.mark.parametrize("suffix", [".py", ".txt", ".md", ".h", ".cpp", ".hpp", ".class"])
@patch("subprocess.run")
def test_is_binary_none_text_files(subprocess_mock: Mock, suffix: str):
    with patch("pathlib.Path.suffix", new_callable=PropertyMock) as mocked_suffix:
        mocked_suffix.return_value = suffix
        path = pathlib.Path(f"/path/to/file{suffix}")
        result = is_binary(path)

        assert result == BinaryType.NONE
        mocked_suffix.assert_called()
        subprocess_mock.assert_not_called()


@pytest.mark.parametrize("suffix", [".a", ".o"])
@patch("subprocess.run")
def test_is_binary_ignore_static_archives(subprocess_mock, suffix: str):
    with patch("pathlib.Path.suffix", new_callable=PropertyMock) as mocked_suffix:
        mocked_suffix.return_value = suffix
        path = pathlib.Path(f"/path/to/file{suffix}")
        result = is_binary(path)

        assert result == BinaryType.NONE
        mocked_suffix.assert_called_once()
        subprocess_mock.assert_not_called()



import json
import logging
import subprocess
import time
from pathlib import Path

LOGGER = logging.getLogger()


def compress(app_bundle: Path) -> Path:
    tosign_zip = app_bundle.parent / app_bundle.name.replace(".app", "-tosign.zip")
    LOGGER.info(f"Compressing {app_bundle} -> {tosign_zip}")
    subprocess.check_call(["/usr/bin/ditto", "-v", "-c", "-k", "--keepParent", str(app_bundle), str(tosign_zip)])
    return tosign_zip


def remove_apple_double(app_bundle: Path):
    LOGGER.info(f"Removing apple double in {app_bundle}.")
    subprocess.check_call(["find", str(app_bundle), "-type", "f", "-name", "._*", "-delete"])


def submit(tosign_zip: Path, keychain_profile: str, keychain: Path, apple_id: str, team_id: str) -> str:
    """
    {"path":"...","message":"...","id":"..."}
    """
    args = [
        "xcrun",
        "notarytool",
        "submit",
        "--output-format",
        "json",
        "--keychain-profile",
        keychain_profile,
        "--keychain",
        str(keychain),
        "--apple-id",
        apple_id,
        "--team-id",
        team_id,
        str(tosign_zip),
    ]
    LOGGER.info(f"Submitting {tosign_zip}")
    output = json.loads(subprocess.check_output(args))
    LOGGER.info(f"{tosign_zip} submitted successfully with submission id: {output['id']}")
    return output["id"]


def check(submission_id: str, keychain_profile: str, keychain: Path, apple_id: str, team_id: str) -> str:
    """
    {"status":"Accepted","message":"Successfully received submission info","name":"ilastik-1.4.1rc2-arm64-OSX-tosign.zip","id":"cdd139bf-5f03-412e-9f26-dc640f19bd17","createdDate":"2025-02-27T07:25:01.041Z"}

    """
    args = [
        "xcrun",
        "notarytool",
        "info",
        "--output-format",
        "json",
        "--keychain-profile",
        keychain_profile,
        "--keychain",
        str(keychain),
        "--apple-id",
        apple_id,
        "--team-id",
        team_id,
        submission_id,
    ]
    output = json.loads(subprocess.check_output(args))
    LOGGER.info(f"{output['name']} submission id: {output['id']} - status: {output['status']}")
    return output["status"].lower()


def staple(app_bundle: Path):
    args = ["xcrun", "stapler", "staple", str(app_bundle)]
    LOGGER.info(f"Stapling {app_bundle}")
    output = subprocess.check_output(args)
    LOGGER.info(output)


def notarize_impl(app_path: Path, keychain_profile: str, keychain: Path, apple_id_email: str, team_id: str) -> int:
    """Notarize an .app bundle with given credentials, wait for completion and staple

    This is equivalent to doing the following steps manually:

    # pack to get ready for notarization
    /usr/bin/ditto -v -c -k --keepParent \
        myapp.app myapp-tosign.zip

    # send off to apple:
    xcrun notarytool submit \
        --keychain-profile <your-keychain-profile> \
        --keychain <path-to-keychain> \
        --apple-id  <email-address-of-dev-account@provider.ext> \
        --team-id <your-team-id> \
        "myapp-tosign.zip"

    # wait for notarization is complete
    xcrun notarytool wait \
        --keychain-profile <your-keychain-profile> \
        --keychain <path-to-keychain> \
        --apple-id  <email-address-of-dev-account@provider.ext> \
        --team-id <your-team-id> \
        <notarization-request-id>

    # once this is done, staple:
    xcrun stapler staple myapp.app

    """
    remove_apple_double(app_path)
    tosign_zip = compress(app_path)
    submission_id = submit(tosign_zip, keychain_profile, keychain, apple_id_email, team_id)

    OVERALL_TIMEOUT = 40 * 60
    SLEEP_S = 60
    timeout = time.perf_counter() + OVERALL_TIMEOUT

    status = "NEVER CHECKED"
    while timeout > time.perf_counter():

        status = check(submission_id, keychain_profile, keychain, apple_id_email, team_id)
        LOGGER.info(f"Submission status {status} for {submission_id}")
        if status == "accepted":
            break
        time.sleep(SLEEP_S)

    LOGGER.info(f"Notarization finished with {status=}")
    if status == "accepted":
        staple(app_path)

    if status == "accepted":
        return 0
    else:
        return -1



import re
import subprocess
from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path
from typing import Optional

from packaging import version

from ._commands import Command
from ._util import BinaryObj, run_logged

_LOAD_DYLIB_REGEX = re.compile(r"\s*name (?P<dylib>.+) \(offset \d+\)$")
_LOAD_RCPATH_REGEX = re.compile(r"\s*path (?P<rc_path>.+) \(offset \d+\)$")
_LOAD_COMMAND_REGEX = re.compile(r"(Load command \d+.*?)(?=Load command \d+|$)", re.DOTALL)


_VTOOL_OUT_PARSE_PLATFORM = re.compile(r"platform (?P<platform>.*)\n", re.IGNORECASE)
_VTOOL_OUT_PARSE_MINOS = re.compile(r"minos (?P<minos>.*)\n", re.IGNORECASE)
_VTOOL_OUT_PARSE_SDK = re.compile(r"sdk (?P<sdk>.*)\n", re.IGNORECASE)
_VALID_VER = re.compile(r"\d+\..*", re.IGNORECASE)


class DependencyNotFountInBundle(Exception):
    pass


@dataclass
class LoadCommand:
    index: str
    cmd: str
    cmd_size: str
    cmd_specifics: list[str]

    @staticmethod
    def from_otool_output(otool_l_output) -> "LoadCommand":
        lines: list[str] = [x.strip() for x in otool_l_output.split("\n")]

        # strip trailing empty lines
        lines = [line for line in lines if line]

        index = re.search(r"Load command (\d+)", lines[0]).groups()[0]
        cmd = re.search(r"cmd (\S+)", lines[1]).groups()[0]
        cmd_size = re.search(r"cmdsize (\d+)", lines[2]).groups()[0]

        additional = lines[3:]
        return LoadCommand(index=index, cmd=cmd, cmd_size=cmd_size, cmd_specifics=additional)


class FILETYPE(IntEnum):
    """
    ref: https://github.com/apple/darwin-xnu/blob/main/EXTERNAL_HEADERS/mach-o/loader.h
    """

    relocatable_object_file = 1
    demand_paged_executable_file = 2
    fixed_VM_shared_library_file = 3
    core_file = 4
    preloaded_executable_file = 5
    dynamically_bound_shared_library = 6
    dynamic_link_editor = 7
    dynamically_bound_bundle_file = 8
    shared_library_stub_for_static = 9

    companion_file_with_only_debug = 10
    x86_64_kexts = 11
    set_of_mach_o_s = 12

    @staticmethod
    def from_hex_str(hex_string: str) -> "FILETYPE":
        return FILETYPE(int(hex_string, base=16))

    def __repr__(self):
        return f"{self.name}"


@dataclass
class MachOHeader:

    magic: str
    filetype: FILETYPE

    @staticmethod
    def from_otool_output(otool_L_output):
        out = otool_L_output.split("\n")
        line = out[-2].strip().split()
        vals = line
        return MachOHeader(vals[0], FILETYPE.from_hex_str(vals[4]))


@dataclass(frozen=True)
class Build:
    platform: str
    minos: str
    sdk: str

    @staticmethod
    def from_vtool_output(vtool_str: str) -> "Build":
        if m := _VTOOL_OUT_PARSE_PLATFORM.search(vtool_str):
            platform = m.groupdict()["platform"].lower()
        else:
            platform = ""

        if m := _VTOOL_OUT_PARSE_MINOS.search(vtool_str):
            minos = m.groupdict()["minos"]
        else:
            minos = ""

        if m := _VTOOL_OUT_PARSE_SDK.search(vtool_str):
            sdk = m.groupdict()["sdk"]
        else:
            sdk = ""

        return Build(platform=platform, minos=minos, sdk=sdk)

    @staticmethod
    def _version_req_met(v: str) -> bool:
        _MIN_V = version.parse("10.9")
        if version.parse(v) >= _MIN_V:
            return True

        return False

    @property
    def is_valid(self) -> bool:
        """Will pass gatekeeper"""
        if self.platform and _VALID_VER.match(self.minos) and _VALID_VER.match(self.sdk):
            # older versions can not pass gatekeeper
            if Build._version_req_met(self.minos) and Build._version_req_met(self.sdk):
                return True

        return False

    @property
    def can_fix(self) -> bool:
        """Libraries that were build with too old of an sdk can not be fixed."""
        if _VALID_VER.match(self.minos):
            if not Build._version_req_met(self.minos):
                return False

        if _VALID_VER.match(self.sdk):
            if not Build._version_req_met(self.sdk):
                return False

        return True

    @property
    def invalid_fields(self) -> dict[str, str]:
        ret = {}
        if not self.platform:
            ret["platform"] = self.platform

        if not self.sdk:
            ret["sdk"] = self.sdk

        if not self.minos:
            ret["minos"] = self.minos

        return ret

    @property
    def invalid_field_names(self) -> list[str]:
        invalid_fields = self.invalid_fields
        return list(invalid_fields.keys())

    def valid_build(self, default_build: "Build", overwrite=False) -> "Build":
        platform = self.platform
        minos = self.minos
        sdk = self.sdk

        if overwrite:
            return Build(
                platform=default_build.platform or platform,
                minos=default_build.minos or minos,
                sdk=default_build.sdk or sdk,
            )

        if not _VALID_VER.match(sdk) and not _VALID_VER.match(minos):
            assert _VALID_VER.match(default_build.sdk) and _VALID_VER.match(default_build.minos)
            sdk = default_build.sdk
            minos = default_build.minos

        if not _VALID_VER.match(self.sdk):
            if _VALID_VER.match(default_build.sdk):
                sdk_version = version.parse(default_build.sdk)
                # don't set sdk version lower than minos
                minos_version = version.parse(minos)
                sdk = default_build.sdk if sdk_version > minos_version else minos
            else:
                sdk = minos

        if not _VALID_VER.match(minos):
            if _VALID_VER.match(default_build.minos):
                minos_version = version.parse(default_build.minos)
                # don't set minos higher than sdk
                sdk_version = version.parse(sdk)
                minos = default_build.minos if minos_version < sdk_version else sdk
            else:
                minos = sdk

        if not platform:
            platform = default_build.platform

        return Build(platform=platform, minos=minos, sdk=sdk)


def vtool_read(path: Path) -> Build:
    """
    vtool -show-build  ilastik-1.4.1rc2-OSX.app/Contents/ilastik-release/lib/libxcb.1.dylib

    ilastik-1.4.1rc2-OSX.app/Contents/ilastik-release/lib/libxcb.1.dylib:
    Load command 16
          cmd LC_BUILD_VERSION
      cmdsize 24
     platform MACOS
        minos 10.9
          sdk 10.9
       ntools 0
    """
    return Build.from_vtool_output(run_logged(Command(["/usr/bin/vtool", "-show-build", str(path)])))


def vtool_overwrite(path: Path, build: Build) -> Command:
    cmd = Command(
        args=[
            "/usr/bin/vtool",
            "-set-build-version",
            build.platform,
            build.minos,
            build.sdk,
            "-replace",
            "-output",
            str(path),
            str(path),
        ],
    )

    return cmd


@dataclass
class MachOBinary(BinaryObj):
    header: MachOHeader
    rpaths: list[Path]
    dylibs: list[Path]
    build: Optional[Build]
    id_: Optional[Path]


def otool_l(path: Path) -> tuple[LoadCommand, ...]:
    out = run_logged(Command(args=["otool", "-l", str(path)]))
    cmds = tuple(LoadCommand.from_otool_output(x) for x in _LOAD_COMMAND_REGEX.findall(out))
    return cmds


def otool_h(path: Path) -> MachOHeader:
    try:
        out = run_logged(Command(["otool", "-h", str(path)]))
    except subprocess.CalledProcessError:
        return False
    return MachOHeader.from_otool_output(out)


def rpaths(cmds: tuple[LoadCommand, ...]) -> list[Path]:
    rcpath_cmds = [cmd for cmd in cmds if cmd.cmd == "LC_RPATH"]
    paths = []
    for rcpath_cmd in rcpath_cmds:
        p = [x for x in rcpath_cmd.cmd_specifics if x.split()[0] == "path"]
        assert len(p) == 1
        paths.append(Path(_LOAD_RCPATH_REGEX.match(p[0]).groupdict()["rc_path"]))

    return paths


def libid(cmds: tuple[LoadCommand, ...]) -> Optional[Path]:
    id_commands = [cmd for cmd in cmds if cmd.cmd == "LC_ID_DYLIB"]
    if len(id_commands) == 0:
        return None
    id_command = id_commands[0]
    p = [x for x in id_command.cmd_specifics if x.split()[0] == "name"]
    assert len(p) == 1
    if m := _LOAD_DYLIB_REGEX.match(p[0]):
        p = m.groupdict()["dylib"]
        dylib = Path(p)
    else:
        raise ValueError(f"Could not parse command {id_command}")

    return dylib


def dylibs(cmds: tuple[LoadCommand, ...]) -> list[Path]:
    """LC_LOAD_DYLIB

    returns a list of dynamic libraries loaded via load commands
    """
    dylib_cmds = [cmd for cmd in cmds if cmd.cmd in ("LC_LOAD_DYLIB", "LC_REEXPORT_DYLIB")]
    dylibs = []
    for rcpath_cmd in dylib_cmds:
        p = [x for x in rcpath_cmd.cmd_specifics if x.split()[0] == "name"]
        assert len(p) == 1
        if m := _LOAD_DYLIB_REGEX.match(p[0]):
            p = m.groupdict()["dylib"]
            dylibs.append(Path(p))

    return dylibs


def parse_macho(some_path: Path):
    if not some_path.is_absolute():
        some_path = some_path.resolve()
    try:
        header = otool_h(some_path)
        cmds = otool_l(some_path)
        paths = rpaths(cmds)
        lib_id = libid(cmds)
        libs = dylibs(cmds)
    except Exception as e:
        raise ValueError(f"Problem parsing {some_path}") from e

    # Found these to contain multiple architectures and binaries
    # vtool (currently) doesn't seem to handle those.
    # TODO: should warn
    # if header.filetype == FILETYPE.fixed_VM_shared_library_file:
    #     build = None
    # else:
    build = vtool_read(some_path)
    return MachOBinary(some_path, header, paths, libs, build, lib_id)


def fix_lib_id(library_path: Path, new_path: Path) -> Command:
    args = ["install_name_tool", "-id", str(new_path), str(library_path)]
    return Command(args=args)


def fix_load_path(library_path: Path, dependency: Path, new_path: Path) -> Command:
    args = ["install_name_tool", "-change", str(dependency), str(new_path), str(library_path)]
    return Command(args=args)


def remove_rpath(library_path, rpath) -> Command:
    args = ["install_name_tool", "-delete_rpath", str(rpath), str(library_path)]
    return Command(args=args)


def fix_rpath(library_path, old_rpath, new_rpath) -> Command:
    args = ["install_name_tool", "-rpath", str(old_rpath), str(new_rpath), str(library_path)]
    return Command(args=args)


def sign_impl(entitlement_file: Path, developer_id: str, path: Path) -> Command:
    args = [
        "/usr/bin/codesign",
        "--entitlements",
        str(entitlement_file),
        "--timestamp",
        "-o",
        "runtime",
        "-f",
        "-s",
        developer_id,
        str(path),
    ]

    return Command(args=args, retry_backoff=True)
